#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Warning: this reallocation script is deprecated and will be move to deforestfoot.


Reallocate import quantities to the original country of production of a commodity.

When attempting to attribute a land footprint of import to a particular
country, the country of import is not necessarily the country of production.
For example if country A imports a commodity from country B and country B has
itself imported that commodity from country C. You don't want to directly
attribute country A's land footprint to country B. The purpose of this script
is to reallocate A's footprint to country C. It is done by:

1. Splitting the imported quantity from B into a share produced in B and a
   share imported by B (from 0% to 100%).

2. Then the share imported by B is reallocated to all its trade partners
   according to the share of trade coming from those countries into B. Quantities
   are expressed in weight equivalent of the primary commodity. The conversion to
   a footprint in terms of area occurs at a later stage.

In particular, if there is no production of the crop in country B, then it
makes sense for that production to be fully reattributed to country C.

Example use: see obs3df_methods trade_share_reallocation.py

TODO: hard code the conversion factors and value share for a small list of
      selected products. Then add a reproducible example here.

"""

from typing import Tuple
import warnings
import pandas
import numpy as np

msg = "This script is deprecated and will be moved to the deforestfoot package.\n"
msg += "Update your import path as follows:\n"
msg += "from deforestfoot.common.reallocate import reallocate"
warnings.warn(msg)


def code_columns(cols: list):
    """Given a list of column names, provide corresponding code columns such as
    reporter_code, partner_code or product_code."""
    cols_to_check = ["reporter", "partner", "product", "primary_product"]
    cols_present = [col for col in cols_to_check if col in cols]
    cols_with_index = [col + "_code" for col in cols_present]
    return cols_with_index


def compute_share_prod_imp(
    df_prod: pandas.DataFrame, df_trade: pandas.DataFrame
) -> pandas.Series:
    """Compute the share between production and import for the given list of
    products. This function works on input data frames in wide format.


    If df_trade contains the primary commodity equivalent of all products (i.e.
    I_primcrop + I_productseq), then this will compute  the following share:

                          P_primcrop
            ----------------------------------------
            (P_primcrop + I_primcrop + I_productseq)

    If df_trade contains only the trade in primary commodity (for example
    crops) then this will compute the following share:

                    P_primcrop
            --------------------------
            (P_primcrop + I_primcrop)

    It performs the following steps:

        1. Aggregate the trade data frame by reporters, year and product
        2. Merge with the production data frame
        3. Compute the share

    Returns a copy of the df_prod data frame with import and share_prod_imp columns.
    """
    index = ["reporter", "product", "year"]
    for name, this_df in {"df_prod": df_prod, "df_trade": df_trade}.items():
        for col in index:
            if col not in this_df:
                msg = f"{col} column not found in DataFrame {name}: {this_df.columns}"
                raise KeyError(msg)
    # Add optional code columns only if they are present in both df
    index += [
        col
        for col in code_columns(index)
        if col in df_trade.columns and col in df_prod.columns
    ]
    df_trade_agg = df_trade.groupby(index)["import_quantity"].agg("sum").reset_index()
    df = df_prod.merge(df_trade_agg, on=index, how="outer")
    df.fillna({"import_quantity": 0, "production": 0}, inplace=True)
    df["share_prod_imp"] = df["production"] / (df["import_quantity"] + df["production"])
    return df


def split_prod_imp(
    df: pandas.DataFrame,
    df_share: pandas.DataFrame,
    step: int,
) -> Tuple[pandas.Series, pandas.Series]:
    """Split a quantity between what is produced domestically and what is imported
    The input data frame must contain the share used for splitting."""
    # Depending on whether df contains production or trade data,
    # Join on reporter or partner
    if step == 1 and "partner" not in df.columns:
        rep_or_par = "reporter"
    else:
        rep_or_par = f"partner_{step-1}"
        df_share = df_share.copy()
        df_share.rename(columns={"reporter": rep_or_par}, inplace=True)
    index = [rep_or_par, "primary_product", "year"]
    index += [col for col in code_columns(index) if col in df.columns]
    df_output = df.merge(df_share[index + ["share_prod_imp"]], on=index, how="left")
    primary_eq_var = f"primary_eq_{step - 1}"

    # Debug start
    # if step == 2:
    #     breakpoint()
    #     df_output_fc  = df_output.query("reporter == 'France' and partner_1 == 'Chile' and year == 2018 and product == 'sunflower_seed_oil_crude'")
    #     df_output_fc[["primary_eq_1", "share_prod_imp"]]
    #     df.query("reporter == 'France' and partner_1 == 'Chile' and year == 2018 and product == 'sunflower_seed_oil_crude'")[["primary_eq_1", "prod", "imp"]]
    # # Debug end

    df_output[f"primary_eq_prod_{step}"] = (
        df_output[primary_eq_var] * df_output["share_prod_imp"]
    )
    df_output[f"primary_eq_imp_{step}"] = df_output[primary_eq_var] * (
        1 - df_output["share_prod_imp"]
    )
    # Remove the share
    df_output.drop(columns="share_prod_imp", inplace=True)
    return df_output


def compute_share_by_partners(
    df: pandas.DataFrame, value=None, threshold=None
) -> pandas.DataFrame:
    """Compute the share of import between different trade partners

    param threshold: Threshold in tons below which the rows will be eliminated from the computation

    """
    if value is None:
        value = "import_quantity"
    if threshold is None:
        threshold = 100
    index = ["reporter", "primary_product", "year"]
    index += [col for col in code_columns(index) if col in df.columns]
    # Compute the proportion among all trade partners
    locator = df[value] > threshold
    df_out = df.loc[locator].copy()
    df_out["imp_share_by_p"] = df_out.groupby(index)[value].transform(
        lambda x: x / x.sum()
    )
    return df_out


def allocate_by_partners(
    df_prod: pandas.DataFrame,
    df_trade: pandas.DataFrame,
    step: int,
) -> pandas.DataFrame:
    """Reallocate a quantity, by splitting it between different trade partners"""
    df_trade = df_trade.copy()  # We don't change the input
    # Merge index with optional code columns
    index = ["primary_product", "year"]
    if step == 1:
        index += ["reporter"]
    else:
        index += [f"partner_{step-1}"]
        df_trade.rename(columns={"reporter": f"partner_{step-1}"}, inplace=True)
    index += [col for col in code_columns(index) if col in df_trade.columns]
    df_trade.rename(columns={"partner": f"partner_{step}"}, inplace=True)
    selected_cols = index + [f"partner_{step}", "imp_share_by_p"]
    df = df_prod.merge(df_trade[selected_cols], on=index, how="left")
    var_alloc = f"primary_eq_imp_alloc_{step}"
    var_agg = f"primary_eq_imp_{step}"
    # Reallocate import quantities to partners according to the proportion
    df[var_alloc] = df[var_agg] * df["imp_share_by_p"]
    return df


def aggregate_on_partners(df: pandas.DataFrame, step: int) -> pandas.DataFrame:
    """Aggregate on partners, followed by a rename of reporter to partner.
    Since we are interested in how the partner production is distributed
    Aggregate import and keep track of production related variables.
    Aggregates the previous step: step - 1, so that the step argument
    is consistent with the other functions in this time step.
    """
    index = ["partner", "product", "primary_product", "year"]
    index += [col for col in code_columns(index) if col in df.columns]
    df_agg = (
        df.groupby(index)[f"primary_eq_imp_alloc_{step - 1}"]
        .agg("sum")
        .reset_index()
        .rename(
            columns={
                "partner": "reporter",
                "partner_code": "reporter_code",
                f"primary_eq_imp_alloc_{step - 1}": f"primary_eq_{step - 1}",
            }
        )
    )
    return df_agg


def reallocate(
    df: pandas.DataFrame,
    prim_prod: pandas.DataFrame,
    prim_trade: pandas.DataFrame,
    n_steps: int,
    threshold: float = None,
):
    """Perform all steps of the reallocation of secondary product production in
    primary commodity equivalent

    TODO: define an object at the level of the primary crop (soybans and
    products, palm and related products, chocolate and related products) Which
    has methods to perform the reallocation.

    The `df` argument contains the production or trade flows to be reallocated,
    values expressed in terms of primary crop equivalent. The arguments
    `prim_prod` and `prim_trade` argument are just used to compute the share of
    primary crop production and import and to build the trade reallocation
    matrix for the primary crop.

    If we use comtrade data, it should be both for the df and prim_trade
    arguments. Note the prim_prod argument will necessarily be from FAOSTAT
    because it's the only source that provides this information.

    param: prim_prod, crop production (for example sunflower seeds production)
    param: prim_trade, crop trade (for example sunflower seeds trade)
    param: df, production of processed products (sunflower oil production)
    param: n_steps number of steps

    Returns a dictionary with trade and production for each step.
    """
    if threshold is None:
        threshold = 1
    # TODO: move this to a separate function that does the reallocation of
    # production only Keep a function that does the reallocation of trade only
    product_to_primary_product = {
        "product_code": "primary_product_code",
        "product": "primary_product",
    }
    # Crop production with import and share of production and import
    prod_share = compute_share_prod_imp(prim_prod, prim_trade)
    prod_share.rename(columns=product_to_primary_product, inplace=True)
    trade = prim_trade.rename(columns=product_to_primary_product).copy()
    trade = compute_share_by_partners(trade)
    real = dict()
    # Drop these columns to avoid having them added at each step
    trade.drop(
        columns=["import_quantity", "reporter_code", "partner_code"], inplace=True
    )
    # First step
    df = df.copy()
    # Drop code columns because renaming of partner to reporter is needed at some point
    # and we don't want to rename also partner_code to reporter_code.
    df.drop(columns=["reporter_code"], inplace=True, errors="ignore")
    # Depending on whether df contains production data or trade data
    # add a partner_0 column
    if "partner" not in df.columns:
        df["partner_0"] = np.nan
    else:
        df["partner_0"] = df["partner"]
    # Split production and import
    df = split_prod_imp(df, prod_share, 1)
    real[("prod", 1)] = df
    # Allocate by partners
    real[("trade", 1)] = allocate_by_partners(df_prod=df, df_trade=trade, step=1)
    nrows_trade = len(real[("trade", 1)])
    msg = f"Reallocation step: {1}, production rows:{len(df)}, "
    msg += f"trade rows: {nrows_trade}"
    print(msg)
    # Subsequent steps
    for step in range(2, n_steps + 1):
        df = real[("trade", step - 1)]
        # Keep only rows above threshold
        selector = df[f"primary_eq_imp_alloc_{step - 1}"] > threshold
        df = df.loc[selector].copy()
        # Temporary column needed by the split_prod_imp function
        df[f"primary_eq_{step - 1 }"] = df[f"primary_eq_imp_alloc_{step - 1 }"]
        df.drop(columns="imp_share_by_p", inplace=True)
        # Split production and import
        df = split_prod_imp(df, prod_share, step)
        # Remove temporary column
        df.drop(columns=f"primary_eq_{step - 1 }")
        real[("prod", step)] = df
        # Allocate by partners
        real[("trade", step)] = allocate_by_partners(
            df_prod=df, df_trade=trade, step=step
        )
        nrows_trade = len(real[("trade", step)])
        msg = f"Reallocation step: {step}, production rows:{len(df)}, "
        msg += f"trade rows: {nrows_trade}"
        print(msg)
    # Return
    return real


def merge_reallocated(
    real: dict, rtol: float = None, atol: float = None, check_ignores_na: bool = False
):
    """
    Merge reallocated imports produced in the importing country.

    At each step, look at the production import split and keep only the import
    that was actually produced in each partner country.

    In case of production data, step 1 returns only the production in country
    A. Step 2 returns the import from first level partners that are actually
    produced in those countries. Step 3 returns the imports from the second
    level partners which is the quantity imported from the first level partners
    partners but not produced in these countries, because it was reimported
    from the second level partners.
    """
    if rtol is None:
        rtol = 1e-3
    if atol is None:
        atol = 100
    # Add a partner_0 column for the first step
    # Prepare a grouping index for the consistency check
    index = ["reporter", "primary_product", "year"]
    if "partner" not in real[("prod", 1)].columns:
        real[("prod", 1)]["partner_0"] = np.nan
    else:
        real[("prod", 1)]["partner_0"] = real[("prod", 1)]["partner"]
        index += ["partner_0"]
    # Assemble the import that was actually produced in each partner country
    df = pandas.DataFrame()
    last_step = pandas.DataFrame(real.keys())[1].max()
    selected_columns = ["primary_product", "year", "reporter", "partner"]
    selected_columns += ["primary_eq", "step"]
    for i in range(1, last_step + 1):
        df_step = real[("prod", i)].copy()
        df_step["partner"] = df_step[f"partner_{i-1}"]
        df_step["primary_eq"] = df_step[f"primary_eq_prod_{i}"]
        intermediate_partners = df_step.columns[
            df_step.columns.str.contains("partner_")
        ]
        intermediate_partners = intermediate_partners.to_list()
        df_step["step"] = i
        df = pandas.concat([df, df_step[selected_columns + intermediate_partners]])
    # Check that the aggregation didn't loose data
    df_agg = df.groupby(index)["primary_eq"].agg("sum").reset_index()
    real_prod_agg = (
        real[("prod", 1)].groupby(index)["primary_eq_0"].agg("sum").reset_index()
    )
    df_check = df_agg.merge(real_prod_agg, on=index, how="outer")
    # Ignore NA values in the check, drop any rows that contain NA value in a column
    if check_ignores_na:
        df_check.dropna(inplace=True)
    selector = np.isclose(
        df_check["primary_eq"], df_check["primary_eq_0"], rtol=rtol, atol=atol
    )
    if not all(selector):
        msg = f"With a relative tolerance of {rtol} "
        msg += f"and an absolute tolerance of {atol}, "
        msg += "the primary equivalent sum doesn't match for the following rows\n"
        msg += f"{df_check.loc[~selector]}"
        msg += "\n The primary_eq_0 column comes from the input data,"
        msg += " primary_eq is after the reallocation."
        raise ValueError(msg)
    return df


def sum_last_partner(df):
    """Sum up flows for the last partner at each step, to obtain flows
    between a reporter country and its last partners"""
    cols = df.columns
    cols = cols[cols.str.contains("partner_")]
    df["final_partner"] = "---"
    for col in cols:
        selector = df[col].isna()
        df.loc[~selector, "final_partner"] = df.loc[~selector, col]
    index = ["primary_product", "year", "reporter", "final_partner"]
    df_agg = df.groupby(index)["primary_eq"].agg("sum").reset_index()
    return df_agg



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Third party modules
import pandas

# Internal modules
from biotrade.faostat import faostat
from biotrade.faostat.aggregate import agg_trade_eu_row
from biotrade.faostat.convert import convert_to_eq_rwd_level_1


class FaostatCountry:
    """
    This object gives access to the FAOSTAT data for one country.
    To access data for many countries, use the faostat object directly.

    Display FAOSTAT forestry production and trade data for one country:

        >>> from biotrade.common.country import Country
        >>> ukr = Country("Ukraine")
        >>> ukr.faostat.forestry_production
        >>> ukr.faostat.forestry_trade
        >>> ukr.faostat.forestry_trade_mirror
        >>> ukr.faostat.forestry_prod_trade_eqrwd

    Display FAOSTAT crop production and trade data for one country:

        >>> from biotrade.common.country import Country
        >>> bra = Country("Brazil")
        >>> bra.faostat.crop_production
        >>> bra.faostat.crop_trade
        >>> bra.faostat.crop_trade_mirror

    Note: crop trade selection has the longest query time at 6 seconds.

        %timeit bra.faostat.crop_trade
        6.06 s ± 206 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
    """

    def __repr__(self):
        return '%s object name "%s"' % (self.__class__, self.country_name)

    def __init__(self, parent):
        """Get the country name from this object's parent class."""
        self.country_name = parent.country_name

    @property
    def forestry_production(self):
        """FAOSTAT forestry production data for one reporter country"""
        return faostat.db.select(
            table="forestry_production", reporter=self.country_name
        )

    @property
    def forestry_trade(self):
        """FAOSTAT forestry bilateral trade data (trade matrix) for one
        reporter country and all its partner countries"""
        df = faostat.db.select(table="forestry_trade", reporter=self.country_name)
        return df

    @property
    def forestry_trade_eu_row(self):
        """FAOSTAT forestry bilateral trade with partners aggregated by EU and Rest of the World"""
        return agg_trade_eu_row(self.forestry_trade, grouping_side="partner")

    @property
    def forestry_trade_mirror(self):
        """FAOSTAT forestry bilateral trade data (trade matrix) for alls
        reporter countries and one partner country"""
        return faostat.db.select(table="forestry_trade", partner=self.country_name)

    @property
    def forestry_trade_mirror_eu_row(self):
        """FAOSTAT mirror flows with reporters aggregated by EU and ROW"""
        df = agg_trade_eu_row(self.forestry_trade_mirror, grouping_side="reporter")
        # TODO: Rename column with a "mirror" suffix
        # df.rename(columns={
        # In the element column, exports become import and vice versa.
        return df

    @property
    def forestry_prod_trade_eqrwd(self):
        """FAOSTAT forestry production and trade data in volume equivalent roundwood

        Bilateral trade products are not the same as the production ones, they
        need to be aggregated using a mapping table defined in faostat.product.groups

        Load the data frame and show available elements:

            >>> from biotrade.common.country import Country
            >>> ukr = Country("Ukraine")
            >>> pt_eqrwd = ukr.faostat.forestry_prod_trade_eqrwd
            >>> print(pt_eqrwd.element.unique())
            >>> print(pt_eqrwd)

        """
        index = [
            "reporter_code",
            "reporter",
            "product_level_1",
            "element",
            "unit",
            "year",
        ]
        # Aggregate trade by production relevant product groups
        ft1_agg = (
            self.forestry_trade_eu_row.merge(
                faostat.products.forestry_trade_groups, on="product"
            )
            .groupby(index)
            .agg(value=("value", sum))
            .reset_index()
            .rename(columns={"product_level_1": "product"})
        )
        # Convert to roundwood equivalent volumes #
        ft1eurow_eqr = convert_to_eq_rwd_level_1(ft1_agg)
        fp1_eqr = convert_to_eq_rwd_level_1(self.forestry_production).drop(
            columns={"element_code", "flag", "period", "product_code"}
        )
        # check that columns are identical
        set(fp1_eqr).symmetric_difference(ft1eurow_eqr)
        # concatenate
        df = pandas.concat([fp1_eqr, ft1eurow_eqr])
        df = df.reset_index()
        return df

    @property
    def crop_production(self):
        """FAOSTAT crop production data for one reporter country"""
        return faostat.db.select(table="crop_production", reporter=self.country_name)

    @property
    def crop_trade(self):
        """FAOSTAT crop bilateral trade data (trade matrix) for one
        reporter country and all its partner countries"""
        df = faostat.db.select(table="crop_trade", reporter=self.country_name)
        return df

    @property
    def crop_trade_eu_row(self):
        """FAOSTAT crop bilateral trade with partners aggregated by EU and Rest of the World"""
        return agg_trade_eu_row(self.crop_trade, grouping_side="partner")

    @property
    def crop_trade_mirror(self):
        """FAOSTAT crop bilateral trade data (trade matrix) for alls
        reporter countries and one partner country"""
        return faostat.db.select(table="crop_trade", partner=self.country_name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Download data from the Eurostat bulk download facility.

Noemi Cazzaniga's package [eurostat](https://pypi.org/project/eurostat/) was
taken as an inspiration to load Eurostat data from the bulk download
repository.

"""
# Standard imports
import logging
import re

# Third party libraries
import numpy as np
import pandas


class Pump:
    """
    Download exchange rates from the Eurostat bulk download platform

        >>> from biotrade.eurostat import eurostat
        >>> exch_rates = eurostat.pump.download_bulk_df("ert_bil_eur_m")

    Download the table of content of all datasets and search for population change

        >>> from biotrade.eurostat import eurostat
        >>> toc = eurostat.pump.download_toc()
        >>> toc.loc[toc.title.str.contains("population change", case=False)]

    """

    # Log debug and error messages
    logger = logging.getLogger("biotrade.eurostat")
    # Base URL to load trade trade data from the API
    url_api_base = "https://ec.europa.eu/eurostat/estat-navtree-portlet-prod/"
    url_bulk_file = url_api_base + "BulkDownloadListing?sort=1&file=data%2F"
    # URL of the table of content
    url_toc = url_api_base + "BulkDownloadListing?sort=1&file=table_of_contents_en.txt"

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        # Mapping table used to rename columns
        self.column_names = self.parent.column_names

    def sanitize_variable_names(self, df, renaming_from, renaming_to):
        """
        Sanitize column names using the mapping table column_names.csv
        Use snake case in column names and replace some symbols
        """
        # Rename columns to snake case
        df.rename(columns=lambda x: re.sub(r"\W+", "_", str(x)).lower(), inplace=True)
        # Map columns using the naming convention defined in self.column_names
        mapping = self.column_names.set_index(renaming_from).to_dict()[renaming_to]
        # Discard nan keys of mapping dictionary
        mapping.pop(np.nan, None)
        # Rename columns using the naming convention defined in self.column_names
        df.rename(columns=mapping, inplace=True)
        return df

    def download_bulk_df(self, code):
        """
        Method of Pump class to download data from the eurostat bulk facility.
        The output is reshaped to wide format.

        The `read_csv` uses many delimiters to deal with the fact that the
        first 3 columns of the csv are comma separated while the rest of the
        data frame is tab separated. In addition, the first tab separation is
        only a tab, while the subsequent separation are a tab and a space.

        :param (str) code, the dataset code
        :output (data frame) in long format with a period column

        Example:

            >>> from biotrade.eurostat import eurostat
            >>> exch_rates = eurostat.pump.download_bulk_df("ert_bil_eur_m")
            >>> demo_gind = eurostat.pump.download_bulk_df("demo_gind")

        """
        # Build the API bulk URL
        url = f"{self.url_bulk_file}{code}.tsv.gz"
        df = pandas.read_csv(url, sep=",|\t| \t", na_values=":", engine="python")
        # Rename columns to snake case
        df.rename(columns=lambda x: re.sub(r"\W+", "_", str(x)).lower(), inplace=True)
        # If there is a time column, reshape to long format
        if any(df.columns.str.contains("time")):
            time_column = df.columns[df.columns.str.contains("time")][-1]
            # Id columns are before the time columns
            id_columns = df.loc[:, :time_column].columns
            df = df.melt(id_vars=id_columns, var_name="period", value_name="value")
            # Remove "\time" from the rightmost column of the index
            df = df.rename(columns=lambda x: re.sub(r"\\time", "", x))
        return df

    def download_toc(self):
        """
        Download the table of content of the eurostat bulk download facility as a data frame

        :output (data frame)

        Example:

            >>> from biotrade.eurostat import eurostat
            >>> toc = eurostat.pump.download_toc()

        Find all datasets that contain the word "population change"

            >>> toc.loc[toc.title.str.contains("population change", case=False)]

        """
        df = pandas.read_csv(self.url_toc, sep="\t")
        # Remove the empty values column
        df.drop(columns="values", inplace=True)
        # Remove white spaces at the beginning and end of strings
        df = df.applymap(lambda x: x.strip() if isinstance(x, str) else x)
        return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

Get the product mapping

Usage:

    >>> from biotrade.common.products import comtrade_faostat_mapping
    >>> from biotrade.common.products import regulation_products

Source  http://datalab.review.fao.org/datalab/caliper/web/classification-page/41
FAOSTAT Commodity List (FCL) "Download mappings" with 2 csv tables:

- http://datalab.review.fao.org/datalab/caliper/web/sites/default/files/2020-01/FCL_HS_mappings_2020-01-07.csv
  which contains FCL codes, labels and the correspondence with Comtrade HS codes.

- http://datalab.review.fao.org/datalab/caliper/web/sites/default/files/2020-01/HS_FCL_mappings_2020-01-07.csv
  which contains Comtrade HS codes, Comtrade HS labels and the correspondence with FCL.

comtrade_faostat_mapping was made from the second table, HS_FCL mapping.


Only the lowest level of FAOSTAT product codes should be entered in this table.
Groups such as the faostat product sawnwood which is an aggregate of sawnwood
non coniferous and sawnwood coniferous should be handled at another level and
only the lowest level product codes of sawnwood non coniferous and sawnwood
coniferous should be present in the mapping table.

A Comtrade product code can only have 2,4 or 6 digits, the code blows checks
that this is the case. For example the code for coffee start with a zero
"090111" (it should not be "90111").

          09                  09 - Coffee, tea, mate and spices
        0901  0901 - Coffee, whether or not roasted or decaf...
      090111      090111 - Coffee; not roasted or decaffeinated

"""

# Third party modules
import pandas
from biotrade import module_dir

config_data_dir = module_dir / "config_data"
# Note the variable name is lower case as suggested in PEP 8
# https://peps.python.org/pep-0008/#function-and-variable-names
comtrade_faostat_mapping = pandas.read_csv(
    config_data_dir / "comtrade_faostat_product_mapping.csv",
    # Force Comtrade to be a string
    dtype={
        "comtrade_code": "str",
        # FAOSTAT code remains an integer
        "faostat_code": "int",
    },
)

regulation_products = pandas.read_csv(
    config_data_dir / "regulation_products.csv",
    # Force Comtrade codes to be strings
    dtype={
        "regulation_code": "str",
        "comtrade_primary_code": "str",
        "hs_4d_code": "str",
        "hs_6d_code": "str",
    },
)

##################################
# Check Comtrade FAOSTAT mapping #
##################################
# A Comtrade product code can only have 2,4 or 6 digits, check this is the case.
nchar = comtrade_faostat_mapping["comtrade_code"].str.len().unique()
if not set(nchar).issubset({2, 4, 6}):
    comtrade_faostat_mapping["nchar"] = comtrade_faostat_mapping[
        "comtrade_code"
    ].str.len()
    msg = "A Comtrade product code can only have 2,4 or 6 digits. "
    msg += "The following codes do not comply:\n"
    msg += f"{comtrade_faostat_mapping.query('not nchar.isin([2,4,6])')}"
    raise ValueError(msg)

# Check duplicates
# One comtrade code should always be associated to one and only one faostat code
# A faostat code can be associated to many comtrade codes
dup_codes = comtrade_faostat_mapping["comtrade_code"].duplicated(keep=False)
if any(dup_codes):
    msg = "Comtrade codes are not unique "
    msg += "in the comtrade to faostat mapping table.\n"
    msg += "The following duplicates are present:\n"
    msg += f"{comtrade_faostat_mapping.loc[dup_codes]}"
    raise ValueError(msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

Copied from https://docs.python.org/3/howto/logging-cookbook.html

Create a biotrade logger with file handler:

    >>> from biotrade.logger import create_logger
    >>> create_logger("biotrade")

This logger can be then used directly using only the logging module itself from
any sub module or script:

    >>> import logging
    >>> logger = logging.getLogger('biotrade.sub_module')
    >>> logger.info("Doing this and that")
"""

# Third party modules
import logging
import os
from pathlib import Path

# Internal modules
from biotrade import data_dir


def create_logger():
    """Create a logger to keep track of debug and error messages"""
    # create logger with 'biotrade'
    logger = logging.getLogger("biotrade")
    # If it exists already it will just be reused
    if logger.hasHandlers():
        return
    # Set level
    logger.setLevel(logging.DEBUG)
    if os.environ.get("BIOTRADE_LOG"):
        path = Path(os.environ["BIOTRADE_LOG"])
    else:
        path = data_dir
    # create file handler which logs even debug messages
    fh = logging.FileHandler(path / "biotrade.log")
    fh.setLevel(logging.DEBUG)
    # create console handler
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    # create formatter and add it to the handlers
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)
    logger.info(
        "Created a logger with file handler %s.", str(path / "biotrade.log")
    )



"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

Script made to export data related to all commodities and tree of the key products, for the web platform

"""


def main():
    import pandas as pd
    from biotrade.faostat import faostat
    from scripts.front_end.functions import (
        save_file,
        main_product_list,
        comtrade_products,
    )

    # Name of product file to retrieve
    faostat_commodity_file = (
        faostat.config_data_dir / "faostat_products_name_code_shortname.csv"
    )
    # Retrieve commodity dataset
    df = pd.read_csv(faostat_commodity_file)
    # Obtain the main product codes of production
    main_products = main_product_list(
        ["crop_production", "forestry_production"]
    )
    # Select only production products
    production_products = df[df.code.isin(main_products)]
    # Filter and rename columns
    column_rename_dict = {
        "code": "product_code",
        "product_short_name_viz": "product_name",
        "commodity_name": "commodity_name",
    }
    production_products = production_products[column_rename_dict.keys()]
    production_products = production_products.rename(columns=column_rename_dict)
    # Save csv files to env variable path or into biotrade data folder
    save_file(production_products, "product_list.csv")
    # Obtain the main product codes of trade
    main_products = main_product_list(["crop_trade"])
    # Select only trade products
    faostat_products = df[df.code.isin(main_products)]
    faostat_products = faostat_products[column_rename_dict.keys()]
    faostat_products = faostat_products.rename(columns=column_rename_dict)
    # Save csv files to env variable path or into biotrade data folder
    save_file(faostat_products, "faostat_product_list.csv")
    # Retrieve regulation products and save the file
    comtrade_products = comtrade_products()
    # Obtain total matches
    rename_col_dict = {"fao_code": "faostat_code", "product_code": "comtrade_code"}
    fao_match = comtrade_products.drop_duplicates(subset="product_code", keep=False)
    fao_match = fao_match[fao_match.match == "total"]
    fao_match = fao_match[rename_col_dict.keys()].rename(columns=rename_col_dict)
    fao_match["faostat_code"] = fao_match.faostat_code.astype(int)
    save_file(fao_match, "mapping_faostat_comtrade_product.csv")
    columns = ["product_code", "product_name", "commodity_name"]
    comtrade_products = comtrade_products.drop_duplicates(subset=columns)[
        columns
    ].reset_index(drop=True)
    save_file(comtrade_products, "comtrade_product_list.csv")


# Needed to avoid running module when imported
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

"""


def value_quantity_share(df, variable, df_relationship, index_list):
    """
    Function to calculate value/quantity share. The value share is necessary
    to compute the land footprint of EU's imports.
    Reference documents: technical report Cuypers 2013 "The impact
    of EU consumption on deforestation: Comprehensive analysis of the impact of
    EU consumption on deforestation"

    :param df, a data frame of production/trade flows
    :param variable, str whether it's the "value_share" or the "quantity_share"
    :param df_relationship, a data frame with a parent column and a child
    column
    :param index_list, a list of columns used as aggregation index, before
    computing the value/quantity share
    :return the aggregate dataframe with the quantity/value share values

    For example compute the quantity share coefficients of soy products
    "oil_soybean" and "cake_soybeans" resulting from "soybeans" primary
    commodity.

    Import dependencies and select bilateral trade of soy

        >>> from biotrade.faostat import faostat
        >>> from biotrade.faostat.share_coefficients import value_quantity_share
        >>> import pandas
        >>> db = faostat.db
        >>> soy = db.select(table="crop_trade", product = "soy")

    Define the variable quantity_share to compute

        >>> variable = "quantity_share"

    Define the index list to obtain aggregate data:

        >>> index = ["reporter", "partner", "product", "year"]

    Define relationship dataframe

        >>> relationship = pandas.DataFrame({
        >>>     'parent': ['soybeans','soybeans'],
        >>>     'child': ['oil_soybean', 'cake_soybeans']})

    Compute the quantity share coefficients and return the dataframe
    containing the corresponding column

        >>> df_quantity_share = value_quantity_share(
        >>>     soy,
        >>>     variable,
        >>>     relationship,
        >>>     index,
        >>> )

    Alternatively the relationship can be obtained from the commodity tree:

        >>> import pandas
        >>> tree = pandas.read_csv(faostat.config_data_dir / "crop_commodity_tree_faostat.csv")
        >>> relationship2 = tree.query("bp==1 and parent=='soybeans'")[["parent","child"]]
        >>> relationship2

    """
    index_list = index_list.copy()

    # select dataframe depending on variable choice
    if variable == "value_share":
        # select df rows where element is export value
        df = df[df["element"] == "export_value"]
    elif variable == "quantity_share":
        # select df rows where element is export quantity
        df = df[df["element"] == "export_quantity"]

    # df grouped by index list
    df_agg = df.groupby(index_list).agg(value=("value", sum)).reset_index()

    # right join to select only rows of df_agg which contains children products
    # of df_relationship
    df_share = df_agg.merge(
        df_relationship,
        how="right",
        left_on="product",
        right_on="child",
    )

    # define new index list based on parent column instead of product
    index_list.remove("product")
    index_list.append("parent")

    # new dataframe grouped by index_list with sum_children column
    # corresponding to the sum of all children products
    df_sum_parent = (
        df_share.groupby(index_list).agg(sum_children=("value", sum)).reset_index()
    )

    # merge the two dataframe based on index_list
    df_share = df_share.merge(
        df_sum_parent,
        how="left",
        left_on=index_list,
        right_on=index_list,
    )

    # add new column computing the value/quantity share as value column
    # divided by the sum children column
    df_share[variable] = (df_share.value / df_share.sum_children).fillna(0)

    return df_share



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC Biomass Project.
Unit D1 Bioeconomy.
"""


# Internal modules
from biotrade.faostat.country import FaostatCountry
from biotrade.hwp.country import HwpCountry


class Country:
    """
    This object gives access to the data pertaining to one country.

    For example display FAOSTAT forestry production data for one country:

        >>> from biotrade.common.country import Country
        >>> ukr = Country("Ukraine")
        >>> ukr.faostat.forestry_production

    FAOSTAT crop trade for one country

        >>> from biotrade.common.country import Country
        >>> bra = Country("Brazil")
        >>> bra.faostat.crop_trade
        >>> bra.faostat.crop_trade_mirror

    """

    def __repr__(self):
        # TODO: add iso 3 code, faostat code and UN Comtrade code here
        return '%s object name "%s"' % (self.__class__, self.country_name)

    def __init__(self, country_name):
        """Initialize a country with its name."""
        self.country_name = country_name

    @property
    def faostat(self):
        """FAOSTAT data for one country"""
        return FaostatCountry(self)

    @property
    def hwp(self):
        """Harvested Wood products mitigation potential"""
        return HwpCountry(self)



"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

Script made to export data related to countries, for the web platform. DEPRECATED

"""

import pandas as pd
from biotrade.faostat import faostat
from functions import save_file

# Name of reporter file to retrieve
reporter_file = faostat.config_data_dir / "faostat_country_groups.csv"
# Retrieve dataset
reporter = pd.read_csv(reporter_file)
# Columns to be retained and to change name
column_dict = {
    "iso3_code": "reporter_code",
    "fao_table_name": "reporter_name",
    "fao_status_info": "status",
}
# Drop rows where iso3 codes are nan
reporter = reporter[column_dict.keys()].rename(columns=column_dict).dropna()
# Define current countries and old ones
selector = reporter.status == "old"
reporter.loc[~selector, "status"] = "iso3"
# Add aggregation reporter code and reporter for average data
reporter = pd.concat(
    [
        reporter,
        pd.DataFrame(
            [
                ["OTH", "Others", "agg"],
                ["EU27", "European Union", "agg"],
                ["ROW", "Rest Of the World", "agg"],
            ],
            columns=reporter.columns,
        ),
    ],
    ignore_index=True,
)
# Remove autonomous regions of France
reporter = reporter[
    ~reporter.reporter_code.isin(
        [
            "GUF",
            "GLP",
            "PYF",
            "SPM",
            "CPT",
            "NCL",
            "MTQ",
            "MYT",
            "MAF",
            "BLM",
            "ATF",
            "REU",
            "WLF",
        ]
    )
].reset_index(drop=True)
# Save csv files to env variable path or into biotrade data folder
save_file(reporter, "reporter_list.csv")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani

Copyright (c) 2023 European Union
Licenced under the MIT licence
"""

# Imports #
from setuptools import setup, find_namespace_packages
from os import path

# Load the contents of the README file #
this_dir = path.abspath(path.dirname(__file__))
readme_path = path.join(this_dir, "README.md")
with open(readme_path, encoding="utf-8") as handle:
    readme = handle.read()

# Call setup #
setup(
    name="biotrade",
    version="0.3.16",
    description="Agriculture and forestry statistics.",
    license="MIT",
    url="https://gitlab.com/bioeconomy/forobs/biotrade/",
    author="Paul Rougieux, Selene Patani",
    author_email="paul.rougieux@gmail.com",
    packages=find_namespace_packages(exclude=["notebooks", "scripts"]),
    install_requires=[
        "pandas",
        "sqlalchemy",
        "sqlalchemy_utils",
        "psycopg2-binary",
        "requests",
        "scipy",
        "pymannkendall",
        "matplotlib",
        "comtradeapicall",
    ],
    extras_require={"api": ["fastapi", "uvicorn"]},
    python_requires=">=3.7",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
    ],
    include_package_data=True,
)



"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

Script made to export data of products and countries (averaging over periods of 5 years except last year of db) related to harvested area/production of the regulation and commodity tree products, for the web platform

"""


def main():
    # Import internal dependencies
    import pandas as pd
    import numpy as np
    from biotrade.faostat import faostat
    from scripts.front_end.functions import COLUMN_PERC_SUFFIX
    from scripts.front_end.functions import (
        filter_production_data,
        aggregated_data,
        main_product_list,
        average_results,
        reporter_iso_codes,
        replace_zero_with_nan_values,
        save_file,
    )

    # Obtain the main product codes
    main_product_list = main_product_list(["crop_production", "forestry_production"])
    # Query db to obtain data
    crop_df = faostat.db.select(
        table="crop_production",
        product_code=main_product_list,
        element=["production", "area_harvested", "stocks", "yield"],
    )
    # Select wood production data
    wood_df = faostat.db.select(
        table="forestry_production",
        product_code=main_product_list,
        element=["production"],
    )
    # Merge data
    df = pd.concat([crop_df, wood_df], ignore_index=True)
    # Filter df with the common most recent year of the several element types
    most_recent_year = df.groupby("element")["year"].max().min()
    df = df[df.year <= most_recent_year]
    # Filter at country level and change units
    df = filter_production_data(df)
    # Aggregate french territories values to France and add them to the dataframe
    code_list = [68, 69, 87, 135, 182, 270, 281]
    agg_country_code = 68
    agg_country_name = faostat.country_groups.df
    agg_country_name = agg_country_name[
        agg_country_name.faost_code == agg_country_code
    ].fao_table_name.values[0]
    df = aggregated_data(df, code_list, agg_country_code, agg_country_name)
    # Substitute faostat codes with iso3 codes
    df = reporter_iso_codes(df)
    # Define the columns and codes for the average calculations
    dict_list = [
        {
            "average_col": "reporter_code",
            "percentage_col": "product_code",
            "threshold_code": -1,
            "index_list_add": [],
        },
        {
            "average_col": "product_code",
            "percentage_col": "reporter_code",
            "threshold_code": "OTH",
            "index_list_add": [],
        },
    ]
    # Calculate the averages and percentages with intervals
    intervals = np.concatenate(
        (
            np.array([0.0]),
            np.array(np.linspace(5, 95, 10) / 100),
            np.array([1.0]),
        )
    )
    df_final = average_results(
        df,
        100,
        dict_list,
        intervals,
    )
    # Columns to keep
    drop_column = [
        "element",
        "reporter_code_avg_value",
        "product_code_percentage",
        "product_code_avg_value",
        "reporter_code_percentage",
    ]
    column_list = df_final.columns.tolist()
    for col in drop_column:
        column_list.remove(col)
    # Define dropna columns
    dropna_col = [col for col in df_final.columns if col.endswith(COLUMN_PERC_SUFFIX)]
    df_final = replace_zero_with_nan_values(df_final, dropna_col)
    df_final = df_final.dropna(subset=dropna_col, how="all")
    # Save csv files to env variable path or into biotrade data folder
    harvested_area = df_final[df_final["element"] == "area_harvested"][column_list]
    save_file(harvested_area, "harvested_area_average.csv")
    production = df_final[df_final["element"].isin(["production", "stocks"])][
        column_list
    ]
    save_file(production, "production_average.csv")
    yields = df_final[df_final["element"] == "yield"][column_list]
    save_file(yields, "yield_average.csv")


# Needed to avoid running module when imported
if __name__ == "__main__":
    main()



"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

Script made to export trends related to harvested area/production of countries associated to the regulation and commodity tree products, for the web platform

"""


def main():
    import sys
    from biotrade.faostat import faostat
    import pandas as pd
    import numpy as np
    from scripts.front_end.functions import (
        aggregated_data,
        main_product_list,
        filter_production_data,
        reporter_iso_codes,
        replace_zero_with_nan_values,
        trend_analysis,
        save_file,
    )

    # Default is using multi process
    multi_process = True
    # If module imported, avoid spawning in Windows
    if __name__ != "__main__":
        if sys.platform.startswith("win"):
            multi_process = False
    # Obtain the main product codes
    main_product_list = main_product_list(["crop_production", "forestry_production"])
    # Select quantities from Faostat db for crop data for all countries (code < 1000)
    crop_data = faostat.db.select(
        table="crop_production",
        product_code=main_product_list,
        element=["production", "area_harvested", "stocks", "yield"],
    )
    # Select wood production data
    wood_data = faostat.db.select(
        table="forestry_production",
        product_code=main_product_list,
        element=["production"],
    )
    # Merge data
    crop_data = pd.concat([crop_data, wood_data], ignore_index=True)
    # Filter crop_data with the common most recent year of the several element types
    most_recent_year = crop_data.groupby("element")["year"].max().min()
    crop_data = crop_data[crop_data.year <= most_recent_year]
    # Filter at country level and change units
    crop_data = filter_production_data(crop_data)
    # Aggregate french territories values to France and add them to the dataframe
    code_list = [68, 69, 87, 135, 182, 270, 281]
    agg_country_code = 68
    agg_country_name = faostat.country_groups.df
    agg_country_name = agg_country_name[
        agg_country_name.faost_code == agg_country_code
    ].fao_table_name.values[0]
    crop_data = aggregated_data(
        crop_data, code_list, agg_country_code, agg_country_name
    )
    # Substitute faostat codes with iso3 codes
    crop_data = reporter_iso_codes(crop_data)
    # Consider data after 1985 to calculate trends of last year
    crop_data = crop_data[crop_data["year"] > 1985]
    # Perform trend analysis
    df = trend_analysis(crop_data, multi_process=multi_process)
    # Columns to be retained
    column_list = [
        "reporter_code",
        "product_code",
        "period_regression",
        "relative_change",
        "absolute_change",
        "mk_slope",
        "mk_significance_flag",
        "unit",
    ]
    # Drop nan values
    dropna_col = ["relative_change", "absolute_change", "mk_slope"]
    df = replace_zero_with_nan_values(df, dropna_col)
    # Put nan to period and significance columns when relative change or slope is 0
    # To avoid inconsistencies during the replace zero with nan
    selector = df.mk_slope.isnull()
    df.loc[selector, "period_regression"] = np.nan
    df.loc[selector, "mk_significance_flag"] = 0
    df = df.dropna(subset=dropna_col, how="all")
    # Harvested area data (only the most common recent year of db)
    most_recent_year = df.groupby("element")["year"].max().min()
    df = df[df["year"] == most_recent_year].reset_index(drop=True)
    df_legend = (
        df[["period", "period_change"]]
        .dropna()
        .drop_duplicates()
        .reset_index(drop=True)
    )
    if len(df_legend) != 1:
        raise ValueError(
            f"Legend for trends can have one and only one row.\nThe actual length is: {len(df_legend)}"
        )
    harvested_area = df[df["element"] == "area_harvested"][column_list]
    # Production data (only the most common recent year of db)
    production = df[df["element"].isin(["production", "stocks"])][column_list]
    # Yield data (only the most common recent year of db)
    yields = df[df["element"] == "yield"][column_list]
    save_file(df_legend, "trends_legend.csv")
    save_file(harvested_area, "harvested_area_trends.csv")
    save_file(production, "production_trends.csv")
    save_file(yields, "yield_trends.csv")


# Needed to avoid running module when imported
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

Usage:

    from biotrade.world_bank import world_bank

"""
# External dependencies
import pandas as pd

# Internal modules
from biotrade import module_dir, data_dir, database_url
from biotrade.world_bank.pump import Pump
from biotrade.world_bank.database import DatabaseWorldBankSqlite
from biotrade.world_bank.database import DatabaseWorldBankPostgresql

# Define a logging mechanism to keep track of errors and debug messages
from biotrade.common.logger import create_logger

create_logger()


class WorldBank:
    """
    Parent to various objects dealing with World Bank data
    """

    # Location of the data
    data_dir = data_dir / "world_bank"

    # Location of module configuration data
    config_data_dir = module_dir / "config_data"

    # Load a mapping table used to rename columns
    df = pd.read_csv(config_data_dir / "column_names.csv")
    # Select only relevant columns and remove incomplete mappings
    df = df[["biotrade", "world_bank"]]
    column_names = df[df.isna().sum(axis=1) == 0]

    def __init__(self):
        if not self.data_dir.exists():
            self.data_dir.mkdir()

    @property
    def pump(self):
        """Load data from the World Bank and read it into data frames"""
        return Pump(self)

    @property
    def db_sqlite(self):
        """Database using the SQLite engine"""
        return DatabaseWorldBankSqlite(self)

    @property
    def db_postgresql(self):
        """Database using the PostGreSQL engine"""
        return DatabaseWorldBankPostgresql(self)

    @property
    def db(self):
        """The generic database can be either a PostGreSQL or a SQLite database
        Depending of the value of the environmental variable
        BIOTRADE_DATABASE_URL. If it's not defined use an SQLite database, otherwise
        use a PostGreSQL database as defined in that URL. Note that the environment
        variables are read at the root of this module's directory. In
        particular BIOTRADE_DATABASE_URL is stored into the DATABASE_URL
        variable.
        """
        if database_url is None:
            return DatabaseWorldBankSqlite(self)
        return DatabaseWorldBankPostgresql(self)


# Make a singleton #
world_bank = WorldBank()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Aggregation functions common to all data sources.

"""

import warnings
from pandas.api.types import is_numeric_dtype
import numpy as np

# Internal modules
from biotrade.faostat import faostat

EU_COUNTRY_NAMES_LIST = faostat.country_groups.eu_country_names


def agg_trade_eu_row(
    df,
    grouping_side="partner",
    drop_index_col=None,
    value_col=None,
    index_side=None,
):
    """
    Aggregate bilateral trade data to eu and row as partners

    :param data frame df: Bilateral trade flows from faostat
    :param str grouping_side: "reporter" or "partner" defines on which side of the
    aggregation index country will be grouped together between EU and rest of
    the world, defaults to partner.
    :param drop_index_col list or str: variables to be dropped from the grouping index
    defaults to ["flag"]
    value_col
    :param value_col list of str: variables to be aggregated, defaults to ["value"]
    :param index_side is deprecated; use grouping_side
    :return bilateral trade flows aggregated by eu and row

    Aggregate over many products in one country

        >>> from biotrade.faostat import faostat
        >>> from biotrade.common.aggregate import agg_trade_eu_row
        >>> ft_can = faostat.db.select(table="forestry_trade",
        >>>                                   reporter=["Canada"])
        >>> ft_can_agg = agg_trade_eu_row(ft_can)

    Aggregate the mirror flows

        >>> ft_can_mirror = faostat.db.select(table="forestry_trade",
        >>>                                          partner=["Canada"])
        >>> ft_can_mirror_agg = agg_trade_eu_row(ft_can_mirror,
                grouping_side="reporter")

    Aggregate two products in many countries on both the reporter and partner sides

        >>> swd = faostat.db.select(table="forestry_trade", product="sawnwood")
        >>> swdagg1 = agg_trade_eu_row(swd, grouping_side="partner")
        >>> swdagg2 = agg_trade_eu_row(swdagg1, grouping_side="reporter")
        >>> swdagg2[["product", "reporter", "partner"]].drop_duplicates()

    When aggregating over partner or reporter groups, the reporter takes the
    name of the reporter group, partner information is kept and can be later
    deleted, if only one partner is present for the purpose of concatenating
    with production data.

        In : ft_can_mirror_agg["reporter"].unique()
        Out: array(['eu', 'row'], dtype=object)

    Select crop trade where products contain the word "soy". Then Aggregate
    Brazil Soy Exports.

        >>> from biotrade.faostat import faostat
        >>> from biotrade.common.aggregate import agg_trade_eu_row
        >>> db = faostat.db
        >>> soy_trade = db.select(table="crop_trade", product = "soy", reporter="Brazil")
        >>> soy_trade_agg = agg_trade_eu_row(soy_trade)

    Select Yearly sawnwood oak trade from Comtrade. Then aggregate over EU and Rest of the World.

        >>> from biotrade.comtrade import comtrade
        >>> from biotrade.common.aggregate import agg_trade_eu_row
        >>> swdoak = comtrade.db.select(table="yearly", product_code="440791")
        >>> # Display interesting columns
        >>> swdoak[["reporter", "partner", "year", 'unit', 'quantity', 'net_weight','trade_value']]
        >>> # Aggregate to EU and ROW
        >>> swdoak_agg = agg_trade_eu_row(swdoak, drop_index_col=["flag"],
        >>>                               value_col=['quantity', 'net_weight','trade_value'])

    Select Yearly soy trade from Comtrade then aggregate over EU and Rest of the World

        >>> from biotrade.comtrade import comtrade
        >>> from biotrade.common.aggregate import agg_trade_eu_row
        >>> df = comtrade.db.select("yearly", product_code_start="1507", partner="Brazil")
        >>> cols_of_interest = ['classification', 'year',
        >>>        'flow_code', 'flow', 'reporter_code',
        >>>        'reporter', 'partner_code', 'partner', 'partner_iso',
        >>>        'product_code', 'unit_code', 'unit', 'quantity', 'net_weight', 'trade_value' ]
        >>> dfeurow = agg_trade_eu_row(df[cols_of_interest],
        >>>                         grouping_side = "reporter",
        >>>                         value_col=['quantity', 'net_weight','trade_value'])

    """
    # Make a copy of the data frame so that it will not be modified in place
    df = df.copy()

    # Default argument values
    if drop_index_col is None:
        drop_index_col = ["flag"]
    if value_col is None:
        value_col = ["value"]
    # Change string arguments to lists
    if isinstance(drop_index_col, str):
        drop_index_col = [drop_index_col]
    if isinstance(value_col, str):
        value_col = [value_col]
    # Deprecate the old name for the argument
    if index_side is not None:
        warnings.warn(
            "index_side is deprecated; use grouping_side", DeprecationWarning, 2
        )
        grouping_side = index_side
    # Restrict values taken by the grouping argument
    if grouping_side not in ["reporter", "partner"]:
        raise ValueError(
            "grouping_side can only take the values 'reporter' or 'partner'"
        )
    # Remove "Total FAO" and "World" rows if present
    selector = df["partner"] != "World"
    if "partner_code" in df.columns:
        if is_numeric_dtype(df["partner_code"]):
            selector = selector & (df["partner_code"] < 1000)
    if any(~selector):
        partner_removed = df.loc[~selector, "partner"].unique()
        warnings.warn(f"Removing {partner_removed} from df")
        df = df[selector].copy()
    # Add EU and ROW groups
    country_group = grouping_side + "_group"
    df[country_group] = "row"
    df[country_group] = df[country_group].where(
        ~df[grouping_side].isin(EU_COUNTRY_NAMES_LIST), "eu"
    )
    # Build the aggregation index based on all columns
    index = df.columns.to_list()
    reporter_and_partner_cols = [
        "reporter_code",
        "reporter",
        "partner_code",
        "partner",
        country_group,
    ]
    # Remove the reporter, partner and grouping column from the index
    # Some grouping columns be added back
    for col in set(drop_index_col + reporter_and_partner_cols + value_col):
        if col in df.columns:
            index.remove(col)
    # The aggregation index depends on the grouping_side
    # Add back the columns that are not on the grouping side
    # Keep the code column only if available in df
    if grouping_side == "partner":
        index = ["reporter_code", "reporter", country_group] + index
        if "reporter_code" not in df.columns:
            index.remove("reporter_code")
    if grouping_side == "reporter":
        index = [country_group, "partner_code", "partner"] + index
        if "partner_code" not in df.columns:
            index.remove("partner_code")
    # Aggregate
    df_agg = df.groupby(index, dropna=False)[value_col].agg("sum").reset_index()
    # When aggregating over partner groups, rename country_group to partner
    if grouping_side == "partner":
        df_agg = df_agg.rename(columns={country_group: "partner"})
    # When aggregating over reporter groups, rename reporter_group to reporter
    if grouping_side == "reporter":
        df_agg = df_agg.rename(columns={country_group: "reporter"})
    # Check that the total value hasn't changed
    np.testing.assert_allclose(
        df_agg[value_col].sum(),
        df[value_col].sum(),
        err_msg=f"The total value sum of the aggregated data {df_agg[value_col].sum()}"
        + f" doesn't match with the sum of the input data frame {df[value_col].sum()}",
    )
    return df_agg


def nlargest(df, value_vars, time_vars=None, agg_groups=None, slice_groups=None, n=10):
    """Return the n largest rows by group sorted by the first value column. The
    value is a yearly average by default, this can be changed by passing
    "period" as the time_vars argument for monthly data for example.

    The aggregation is done in two steps: (1) a sum over the grouping variables
    including time, (2) an average over the grouping variables.

    Rows are sorted by the first of the value_vars and slicing takes the
    first n rows in each slice group.

    :param data frame df
    :param value_vars list or str: name of the value columns to be aggregated
    :param time_vars list or str: name of the time columns, defaults to "year"
    :param agg_groups list or str: name of the grouping columns
    :param slice_groups list or str: name of the slicing columns
    :param n int: number of rows to keep in each group
    :return data frame

    For example the 5 largest sawnwood pine trade partners by flow globally
    over January to June 2021:

        >>> from biotrade.comtrade import comtrade
        >>> from biotrade.common.aggregate import nlargest
        >>> sp = comtrade.db.select("monthly", product_code="440711")
        >>> # Keep data for the first semmester of 2021
        >>> sp2021s1 = sp.query("202101<=period & period<=202106 &partner!='World'")
        >>> nlargest(sp2021s1,
        >>>          value_vars=["trade_value", "net_weight"],
        >>>          agg_groups=["reporter", "partner","flow"],
        >>>          slice_groups=["flow"],
        >>>          n=5)

    The 6 largest sawnwood pine world importers

        >>> spw = comtrade.db.select("monthly", product_code="440711", partner="World")
        >>> nlargest(spw,
        >>>          value_vars=["trade_value", "net_weight"],
        >>>          agg_groups=["reporter", "partner", "flow"],
        >>>          slice_groups="flow",
        >>>          n=6)

    The 6 largest wheat producers over 2010-2020, by element

        >>> from biotrade.faostat import faostat
        >>> from biotrade.common.aggregate import nlargest
        >>> wheat = faostat.db.select("crop_production", product_code=15)
        >>> # Select 10 years and remove aggregate regions
        >>> wheat10 = wheat.query("2010<=period & period<=2020 & reporter_code<1000")
        >>> nlargest(wheat10,
        >>>          value_vars="value",
        >>>          agg_groups="reporter",
        >>>          slice_groups="element")

    The 5 largest cocoa beans exporters to the EU and ROW

        >>> from biotrade.faostat import faostat
        >>> from biotrade.common.aggregate import nlargest
        >>> from biotrade.common.aggregate import agg_trade_eu_row
        >>> cocob_raw = faostat.db.select("crop_trade", product_code=661, period_start=2015)
        >>> cocob_raw = cocob_raw.query("element in ['import_quantity', 'import_value']")
        >>> cocob = agg_trade_eu_row(cocob_raw, grouping_side="reporter")
        >>> # Yearly average over the last 5 years
        >>> nlargest(cocob,
        >>>          value_vars="value",
        >>>          agg_groups=["reporter","partner"],
        >>>          slice_groups="element")
        >>> # Yearly values
        >>> nlargest(cocob,
        >>>          value_vars="value",
        >>>          agg_groups=["reporter","partner"],
        >>>          slice_groups=["element", "year"])

    The 12 largest Polish export destination over the past 5 years (notice the
    absence of the slice group)

        >>> from biotrade.common.aggregate import nlargest
        >>> from biotrade.comtrade import comtrade
        >>> pl = comtrade.db.select("yearly", reporter="Poland", product_code_start="44")
        >>> nlargest(pl.query("year>=2017 and flow=='export'"),
        >>>          value_vars=["net_weight", "trade_value"],
        >>>          agg_groups=["reporter", "partner", "flow"],
        >>>          n=12
        >>>         )

    """
    if time_vars is None:
        time_vars = ["year"]
    # Transform all input parameters that are strings to lists
    if isinstance(time_vars, str):
        time_vars = [time_vars]
    if isinstance(agg_groups, str):
        agg_groups = [agg_groups]
    if isinstance(slice_groups, str):
        slice_groups = [slice_groups]
    if isinstance(value_vars, str):
        value_vars = [value_vars]
    # Make sure slice columns are part of the agg_groups
    if agg_groups is None:
        agg_groups = slice_groups
    if agg_groups is not None and slice_groups is not None:
        agg_groups = list(set(agg_groups).union(set(slice_groups)))
    # Compute the sum within each group for each time period (such as each year)
    # Then compute the average for each group over the years
    if agg_groups is not None:
        df_agg = df.groupby(agg_groups + time_vars)[value_vars].agg("sum")
        df_agg = df_agg.groupby(agg_groups)[value_vars].agg("mean")
    else:
        df_agg = df.groupby(time_vars)[value_vars].agg("sum")
    # Sort rows by ascending slice_groups and descending first value column
    if slice_groups is not None:
        df_slice = (
            df_agg.sort_values(
                slice_groups + [value_vars[0]],
                ascending=[True] * len(slice_groups) + [False],
            )
            .groupby(slice_groups)
            .head(n)
        )
    else:
        df_slice = df_agg.sort_values([value_vars[0]], ascending=False).head(n)
    df_slice = df_slice.reset_index()
    return df_slice


def check_aggregate_sum(df, index):
    """Check that the World aggregate given in this data set corresponds to
    the sum of its constituents

    :param data frame df
    :param index list: list of index columns to be used in the group_by

    Example

        >>> from biotrade.comtrade import comtrade
        >>> from biotrade.common.aggregate import nlargest
        >>> sp = comtrade.db.select("monthly", product_code="440711")
        >>> check_aggregate_sum(sp, partner_side = "World")

    """
    # Extract world aggregate



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test functions from:

    - scripts/front_end

Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence
"""

import pandas as pd
from pandas.testing import assert_frame_equal
from pandas.testing import assert_series_equal
from scripts.front_end.functions import aggregated_data, reporter_iso_codes


# def test_reporter_iso_codes():
#     # Create a sample DataFrame for testing
#     df = pd.DataFrame(
#         {
#             "reporter_code": [351, 41, 214, 2, 15, 3, 3, 3],
#             "partner_code": [4, 6, 7, 9, 10, 3, 351, 41],
#             "value": [100, 200, 300, 400, 500, 600, 700, 800],
#         }
#     )
#     # Perform the transformation using the function
#     transformed_df = reporter_iso_codes(df)
#     # Define the expected output DataFrame
#     expected_df = pd.DataFrame(
#         {
#             "reporter_code": ["CHN", "AFG", "ALB"],
#             "partner_code": ["AND", "ARG", "CHN"],
#             "value": [200, 400, 800],
#         }
#     )
#     # Assert that the transformed DataFrame is equal to the expected DataFrame
#     assert_series_equal(
#         transformed_df["reporter_code"], expected_df["reporter_code"]
#     )
#     assert_series_equal(
#         transformed_df["partner_code"], expected_df["partner_code"]
#     )
#     assert_frame_equal(transformed_df, expected_df)


def test_aggregated_data():
    # Sample input data
    df = pd.DataFrame(
        {
            "source": ["a", "a", "a", "a", "a", "a", "a", "a"],
            "reporter_code": [1, 2, 1, 2, 5, 5, 1, 3],
            "partner_code": [7, 7, 6, 6, 10, 1, 2, 7],
            "reporter": ["A", "B", "A", "B", "E", "E", "A", "C"],
            "partner": ["G", "G", "F", "F", "J", "A", "B", "G"],
            "product_code": [11, 11, 13, 14, 15, 15, 11, 11],
            "product": ["11", "11", "13", "14", "15", "15", "11", "11"],
            "element_code": [16, 16, 18, 18, 20, 20, 16, 16],
            "element": ["b", "b", "c", "c", "d", "d", "b", "b"],
            "year": [2020, 2020, 2021, 2021, 2022, 2022, 2020, 2020],
            "period": [2020, 2020, 2021, 2021, 2022, 2022, 2020, 2020],
            "unit": ["kg", "kg", "kg", "kg", "kg", "kg", "kg", "kg"],
            "value": [100, 200, 300, 400, 500, 600, 700, 800],
        }
    )
    code_list = [1, 2, 3]
    agg_country_code = 99
    agg_country_name = "Aggregated"
    # Expected output
    expected_output = pd.DataFrame(
        {
            "source": ["a", "a", "a", "a", "a"],
            "reporter_code": [5, 99, 99, 99, 5],
            "partner_code": [10, 6, 6, 7, 99],
            "reporter": ["E", "Aggregated", "Aggregated", "Aggregated", "E"],
            "partner": ["J", "F", "F", "G", "Aggregated"],
            "product_code": [15, 13, 14, 11, 15],
            "product": ["15", "13", "14", "11", "15"],
            "element_code": [20, 18, 18, 16, 20],
            "element": ["d", "c", "c", "b", "d"],
            "year": [2022, 2021, 2021, 2020, 2022],
            "period": [2022, 2021, 2021, 2020, 2022],
            "unit": ["kg", "kg", "kg", "kg", "kg"],
            "value": [500, 300, 400, 1100, 600],
        }
    )
    # Sort column values
    expected_output = expected_output.sort_values(
        expected_output.columns.tolist()
    ).reset_index(drop=True)
    # Call the function
    result = aggregated_data(df, code_list, agg_country_code, agg_country_name)
    # Sort column values
    result = result.sort_values(result.columns.tolist()).reset_index(drop=True)
    # Assert the result
    assert_series_equal(
        result["value"],
        expected_output["value"],
    )
    assert_frame_equal(result, expected_output)



"""Common functions suitable for many data sources

For example the following functions are common to two or more data sources:

- `biotrade.common.compare.transform_comtrade_using_faostat_codes` transforms
Comtrade data into a format compatible with FAOSTAT data for comparison
purposes.

- `biotrade.common.aggregate.agg_trade_eu_row` aggregate trade to EU
and Rest of the World on the reporter side, the partner side or both.

- `biotrade.common.aggregate.nlargest` return the n largest producers or n largest
trade partners of a product with. Rows are sorted by the first of the
value_vars and slicing takes the first n rows in each slice group.

- `biotrade.common.update.cron_quarterly` and
`biotrade.common.update.cron_bi_weekly` define cron jobs which will
automatically update datasets every three months or every two weeks.

"""



"""
Written by Selene Patani

Copyright (c) 2023 European Union
Licenced under the MIT licence

Script used to compute averages and percentages of trade quantities related to the main commodities, europe and rest of the world imports

"""


def main():
    from scripts.front_end.functions import (
        main_product_list,
        comtrade_products,
        merge_faostat_comtrade_data,
        filter_trade_data,
        aggregated_data,
        reporter_iso_codes,
        average_results,
        replace_zero_with_nan_values,
        save_file,
    )
    from scripts.front_end.functions import (
        COLUMN_PERC_SUFFIX,
        COLUMN_AVG_SUFFIX,
    )
    from biotrade.faostat import faostat
    from biotrade.faostat.aggregate import agg_trade_eu_row
    import pandas as pd

    # Obtain faostat product codes
    faostat_list = main_product_list(["crop_trade"])
    # Obtain comtrade regulation codes
    comtrade_regulation = comtrade_products()
    # Trade data related to product code list
    trade_data = merge_faostat_comtrade_data(
        faostat_list, comtrade_regulation, aggregate=False
    )
    # Filter and change units
    trade_data = filter_trade_data(trade_data)
    # Aggregate french territories values to France and add them to the dataframe
    code_list = [68, 69, 87, 135, 182, 270, 281]
    agg_country_code = 68
    agg_country_name = faostat.country_groups.df
    agg_country_name = agg_country_name[
        agg_country_name.faost_code == agg_country_code
    ].fao_table_name.values[0]
    trade_data = aggregated_data(
        trade_data, code_list, agg_country_code, agg_country_name
    )
    # Substitute faostat codes with iso3 codes
    trade_data = reporter_iso_codes(trade_data)
    # Define the columns and codes for the average calculations
    dict_list = [
        {
            "average_col": "product_code",
            "percentage_col": "partner_code",
            "threshold_code": "OTH",
            "index_list_add": ["reporter_code"],
        },
        {
            "average_col": "product_code",
            "percentage_col": "reporter_code",
            "threshold_code": "OTH",
            "index_list_add": ["partner_code"],
        },
    ]
    # Consider faostat data
    df_faostat = trade_data[trade_data["source"] == "faostat"]
    # Calculate the averages and percentages
    df_faostat = average_results(
        df_faostat,
        100,
        dict_list,
    )
    # Consider comtrade data (excluded estimated values)
    df_comtrade = trade_data[trade_data["source"] == "comtrade"]
    # Calculate the averages and percentages
    df_comtrade = average_results(
        df_comtrade,
        100,
        dict_list,
    )
    drop_column_list = [
        "element",
        dict_list[1]["percentage_col"] + COLUMN_PERC_SUFFIX,
    ]
    column_list = df_faostat.columns.tolist()
    for drop_column in drop_column_list:
        column_list.remove(drop_column)
    # Define dropna columns
    dropna_col = [
        col
        for col in column_list
        if col.endswith((COLUMN_AVG_SUFFIX, COLUMN_PERC_SUFFIX))
    ]
    # Remove columns not needed for front end
    column_list.remove("product_code_avg_value")
    # Consider selected columns of import quantities and save the file (drop nan)
    df_faostat_avg_imp = df_faostat.copy()
    df_faostat_avg_imp = replace_zero_with_nan_values(df_faostat_avg_imp, dropna_col)
    df_faostat_avg_imp = df_faostat_avg_imp.dropna(subset=dropna_col)
    df_faostat_reporter = df_faostat_avg_imp[
        df_faostat_avg_imp["element"] == "import_quantity"
    ][column_list]
    save_file(df_faostat_reporter, "faostat_average.csv")
    # Consider selected columns of import quantities and save the file (drop nan)
    df_comtrade_avg_imp = df_comtrade.copy()
    df_comtrade_avg_imp = replace_zero_with_nan_values(df_comtrade_avg_imp, dropna_col)
    df_comtrade_avg_imp = df_comtrade_avg_imp.dropna(subset=dropna_col)
    df_comtrade_reporter = df_comtrade_avg_imp[
        df_comtrade_avg_imp["element"] == "import_quantity"
    ][column_list]
    save_file(df_comtrade_reporter, "comtrade_average.csv")
    # drop_column_list = [
    #     "element",
    #     dict_list[0]["percentage_col"] + COLUMN_PERC_SUFFIX,
    # ]
    # column_list = df_faostat.columns.tolist()
    # for drop_column in drop_column_list:
    #     column_list.remove(drop_column)
    # # Define dropna columns
    # dropna_col = [
    #     col
    #     for col in column_list
    #     if col.endswith((COLUMN_AVG_SUFFIX, COLUMN_PERC_SUFFIX))
    # ]
    # # Consider selected columns of export quantities and save the file (drop nan) --> mirror flows
    # df_faostat_avg_exp = df_faostat.copy()
    # df_faostat_avg_exp = replace_zero_with_nan_values(df_faostat_avg_exp, dropna_col)
    # df_faostat_avg_exp = df_faostat_avg_exp.dropna(subset=dropna_col)
    # df_faostat_partner = df_faostat_avg_exp[
    #     df_faostat_avg_exp["element"] == "export_quantity"
    # ][column_list]
    # save_file(df_faostat_partner, "faostat_average_mf.csv")
    # # Consider selected columns of export quantities and save the file (drop nan) --> mirror flows
    # df_comtrade_avg_exp = df_comtrade.copy()
    # df_comtrade_avg_exp = replace_zero_with_nan_values(df_comtrade_avg_exp, dropna_col)
    # df_comtrade_avg_exp = df_comtrade_avg_exp.dropna(subset=dropna_col)
    # df_comtrade_partner = df_comtrade_avg_exp[
    #     df_comtrade_avg_exp["element"] == "export_quantity"
    # ][column_list]
    # save_file(df_comtrade_partner, "comtrade_average_mf.csv")
    # Consider averages for EU and rest of the world partners
    # Aggregate data with reporters as eu and row
    df_group_reporter = agg_trade_eu_row(
        trade_data,
        grouping_side="reporter",
        drop_index_col=["flag"],
    )
    # Aggregate data with partners as eu and row
    df_group_partner = agg_trade_eu_row(
        trade_data,
        grouping_side="partner",
        drop_index_col=["flag"],
    )
    # Concatenate in a unique df
    df_group = pd.concat(
        [df_group_reporter, df_group_partner],
        ignore_index=True,
    )
    # Substitute with name and codes of the aggregations for the web platform
    selector = df_group["reporter"] == "eu"
    df_group.loc[selector, "reporter_code"] = "EU27"
    selector = df_group["reporter"] == "row"
    df_group.loc[selector, "reporter_code"] = "ROW"
    selector = df_group["partner"] == "eu"
    df_group.loc[selector, "partner_code"] = "EU27"
    selector = df_group["partner"] == "row"
    df_group.loc[selector, "partner_code"] = "ROW"
    df_group[["reporter", "partner"]] = df_group[["reporter", "partner"]].replace(
        ["eu", "row"], ["European Union", "Rest Of the World"]
    )
    # Consider faostat data
    df_faostat = df_group[df_group["source"] == "faostat"]
    # Calculate the averages and percentages
    df_faostat = average_results(
        df_faostat,
        100,
        dict_list,
    )
    # Consider comtrade data (excluded estimated values)
    df_comtrade = df_group[df_group["source"] == "comtrade"]
    # Calculate the averages and percentages
    df_comtrade = average_results(
        df_comtrade,
        100,
        dict_list,
    )
    drop_column_list = [
        "element",
        dict_list[1]["percentage_col"] + COLUMN_PERC_SUFFIX,
    ]
    column_list = df_faostat.columns.tolist()
    for drop_column in drop_column_list:
        column_list.remove(drop_column)
    # Define dropna columns
    dropna_col = [
        col
        for col in column_list
        if col.endswith((COLUMN_AVG_SUFFIX, COLUMN_PERC_SUFFIX))
    ]
    # Remove columns not needed for front end
    column_list.remove("product_code_avg_value")
    # Consider selected columns of import quantities and save the file (drop nan)
    df_faostat_avg_imp = df_faostat.copy()
    df_faostat_avg_imp = replace_zero_with_nan_values(df_faostat_avg_imp, dropna_col)
    df_faostat_avg_imp = df_faostat_avg_imp.dropna(subset=dropna_col)
    df_faostat_reporter = df_faostat_avg_imp[
        (df_faostat_avg_imp["element"] == "import_quantity")
        & (df_faostat_avg_imp["reporter_code"].isin(["EU27", "ROW"]))
    ][column_list]
    save_file(df_faostat_reporter, "faostat_average_eu_row.csv")
    # Consider selected columns of import quantities and save the file (drop nan)
    df_comtrade_avg_imp = df_comtrade.copy()
    df_comtrade_avg_imp = replace_zero_with_nan_values(df_comtrade_avg_imp, dropna_col)
    df_comtrade_avg_imp = df_comtrade_avg_imp.dropna(subset=dropna_col)
    df_comtrade_reporter = df_comtrade_avg_imp[
        (df_comtrade_avg_imp["element"] == "import_quantity")
        & (df_comtrade_avg_imp["reporter_code"].isin(["EU27", "ROW"]))
    ][column_list]
    save_file(df_comtrade_reporter, "comtrade_average_eu_row.csv")
    # drop_column_list = [
    #     "element",
    #     dict_list[0]["percentage_col"] + COLUMN_PERC_SUFFIX,
    # ]
    # column_list = df_faostat.columns.tolist()
    # for drop_column in drop_column_list:
    #     column_list.remove(drop_column)
    # # Define dropna columns
    # dropna_col = [
    #     col
    #     for col in column_list
    #     if col.endswith((COLUMN_AVG_SUFFIX, COLUMN_PERC_SUFFIX))
    # ]
    # # Consider selected columns of export quantities and save the file (drop nan) --> mirror flows
    # df_faostat_avg_exp = df_faostat.copy()
    # df_faostat_avg_exp = replace_zero_with_nan_values(df_faostat_avg_exp, dropna_col)
    # df_faostat_avg_exp = df_faostat_avg_exp.dropna(subset=dropna_col)
    # df_faostat_partner = df_faostat_avg_exp[
    #     (df_faostat_avg_exp["element"] == "export_quantity")
    #     & (df_faostat_avg_exp["partner_code"].isin(["EU27", "ROW"]))
    # ][column_list]
    # save_file(df_faostat_partner, "faostat_average_eu_row_mf.csv")
    # # Consider selected columns of export quantities and save the file (drop nan) --> mirror flows
    # df_comtrade_avg_exp = df_comtrade.copy()
    # df_comtrade_avg_exp = replace_zero_with_nan_values(df_comtrade_avg_exp, dropna_col)
    # df_comtrade_avg_exp = df_comtrade_avg_exp.dropna(subset=dropna_col)
    # df_comtrade_partner = df_comtrade_avg_exp[
    #     (df_comtrade_avg_exp["element"] == "export_quantity")
    #     & (df_comtrade_avg_exp["partner_code"].isin(["EU27", "ROW"]))
    # ][column_list]
    # save_file(df_comtrade_partner, "comtrade_average_eu_row_mf.csv")


# Needed to avoid running module when imported
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
You can use this object at the ipython console with the following examples.

Download and return a data frame:

    >>> from biotrade.eurostat import eurostat
    >>> exch_rates = eurostat.pump.download_bulk_df("ert_bil_eur_m")
    >>> demo_gind = eurostat.pump.download_bulk_df("demo_gind")

"""
# Standard modules

# Third party modules
import pandas

# Internal modules
from biotrade.eurostat.pump import Pump

# Define a logging mechanism to keep track of errors and debug messages
from biotrade.common.logger import create_logger
from biotrade import module_dir, data_dir

create_logger()


class Eurostat:
    """
    Parent to various objects dealing with data from Eurostat.
    """

    # Location of module configuration data
    config_data_dir = module_dir / "config_data"

    # Load a mapping table used to rename columns
    df = pandas.read_csv(config_data_dir / "column_names.csv")
    # Select only relevant columns and remove incomplete mappings
    df = df[["biotrade", "eurostat_bulk"]]
    column_names = df[df.isna().sum(axis=1) == 0]

    def __init__(self):
        # Location of the data
        self.data_dir = data_dir / "eurostat"
        if not self.data_dir.exists():
            self.data_dir.mkdir()

    @property
    def pump(self):
        """Download data from Comtrade and send it to the database"""
        return Pump(self)


# Make a singleton #
eurostat = Eurostat()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

Get the module internal list of products groups

    >>> from biotrade.faostat import faostat
    >>> print(faostat.products.forestry_production_groups)
    >>> print(faostat.products.forestry_trade_groups)

"""
# Third party modules
import pandas


class Products(object):
    """
    Comtrade product list, with additional information.
    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        # Directories #
        self.config_data_dir = self.parent.config_data_dir

    @property
    def forestry_production_groups(self):
        """FAOSTAT forestry production groups"""
        df = pandas.read_csv(
            self.config_data_dir / "faostat_forestry_production_groups.csv"
        )
        return df

    @property
    def forestry_trade_groups(self):
        """FAOSTAT forestry trade groups"""
        df = pandas.read_csv(self.config_data_dir / "faostat_forestry_trade_groups.csv")
        return df



"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

Script made to export trends related to bilateral trades of countries associated to palm oil, soybeans, cocoa beans, coffee roasted, meat cattle, hides cattle fresh
"""

# Needed for multiprocessing tool, otherwise Windows spawns multiple threads
if __name__ == "__main__":
    from functions import *

    # Selected product codes
    product_list = [236, 238, 257, 657, 661, 867, 919]
    # Trade data related to product code list
    trade_data = merge_faostat_comtrade_data(product_list)
    # China Mainland + Taiwan data
    df_china = consistency_check_china_data(trade_data)
    # Add China data to trade_data (exclude Taiwan data)
    trade_data = pd.concat(
        [
            trade_data[
                ~((trade_data[["reporter_code", "partner_code"]] == 214).any(axis=1))
            ],
            df_china,
        ],
        ignore_index=True,
    )
    # Substitute faostat codes with iso3 codes
    trade_data = reporter_iso_codes(trade_data)
    # Consider data after 1985 to calculate trends of last year (excluded estimated values)
    trade_data = trade_data[
        (trade_data["year"] > 1985) & (trade_data["estimate_flag"] == 0)
    ]
    # Remove unused columns
    trade_data.drop(columns=["product", "element_code"], inplace=True)
    # Perform trend analysis
    df = trend_analysis(trade_data, multi_process=False)
    # Columns to be retained
    column_list = [
        "reporter_code",
        "partner_code",
        "product_code",
        "period",
        "period_change",
        "period_regression",
        "relative_change",
        "absolute_change",
        "mk_slope",
        "mk_significance_flag",
        "unit",
    ]
    # Drop nan values
    dropna_col = ["relative_change", "absolute_change", "mk_slope"]
    df = replace_zero_with_nan_values(df, dropna_col)
    # Put nan to period and significance columns when relative change or slope is 0
    # To avoid inconsistencies during the replace zero with nan
    selector = df.mk_slope.isnull()
    df.loc[selector, "period_regression"] = np.nan
    df.loc[selector, "mk_significance_flag"] = np.nan
    selector = df.relative_change.isnull()
    df.loc[selector, "period_change"] = np.nan
    df = df.dropna(subset=dropna_col, how="all")
    # Define most recent year for faostat and comtrade data
    trade_data_faostat = df[df["source"] == "faostat"]
    most_recent_year_faostat = sorted(trade_data_faostat.year.unique(), reverse=True)[0]
    trade_data_comtrade = df[df["source"] == "comtrade"]
    most_recent_year_comtrade = sorted(trade_data_comtrade.year.unique(), reverse=True)[
        0
    ]
    # Consider selected columns of import quantities and save the file
    trade_data_faostat_reporter = trade_data_faostat[
        (trade_data_faostat["year"] == most_recent_year_faostat)
        & (trade_data_faostat["element"] == "import_quantity")
    ][column_list]
    save_file(trade_data_faostat_reporter, "faostat_trends.csv")
    # Consider selected columns of import quantities and save the file
    trade_data_comtrade_reporter = trade_data_comtrade[
        (trade_data_comtrade["year"] == most_recent_year_comtrade)
        & (trade_data_comtrade["element"] == "import_quantity")
    ][column_list]
    save_file(trade_data_comtrade_reporter, "comtrade_trends.csv")
    # Consider selected columns of export quantities and save the file --> mirror flows
    trade_data_faostat_partner = trade_data_faostat[
        (trade_data_faostat["year"] == most_recent_year_faostat)
        & (trade_data_faostat["element"] == "export_quantity")
    ][column_list]
    save_file(trade_data_faostat_partner, "faostat_trends_mf.csv")
    # Consider selected columns of export quantities and save the file --> mirror flows
    trade_data_comtrade_partner = trade_data_comtrade[
        (trade_data_comtrade["year"] == most_recent_year_comtrade)
        & (trade_data_comtrade["element"] == "export_quantity")
    ][column_list]
    save_file(trade_data_comtrade_partner, "comtrade_trends_mf.csv")
